package gs.datagen.v0

/** Base class for data generators. Note that generators require some input of
  * type `I`. For a specialization that does not require input, please refer to
  * [[Datagen.NoInput]]. The [[Gen]] type alias `Gen[A]` is equivalent to
  * [[Datagen.NoInput]].
  */
abstract class Datagen[A, -I]:
  /** Generate some data, given some input.
    *
    * @param input
    *   The input provided to this generation.
    */
  def generate(input: I): A

  def flatMap[B, I2 <: I](f: A => Datagen[B, I2]): Datagen[B, I2] =
    new Datagen.Defer[B, I2](input => f(generate(input)).generate(input))

object Datagen:

  /** Produce a generator that emits a single, fixed value.
    *
    * @param data
    *   The fixed data to generate.
    */
  def pure[A](data: A): NoInput[A] =
    NoInput.Pure[A](data)

  /** Used internally to implement `map` by producing a `Datagen` based on the
    * evaluation of some input.
    */
  final class Defer[A, -I](
    f: I => A
  ) extends Datagen[A, I]:
    override def generate(input: I): A = f(input)

  /** Specialization of [[Datagen]] that never accepts input. This is used for
    * fixed or randomized data and is the basis of most generators.
    */
  abstract class NoInput[A] extends Datagen[A, Any]:

    /** Generate a value.
      */
    def gen(): A = generate(())

    def flatMap[B](f: A => NoInput[B]): NoInput[B] =
      new NoInput.Defer[B](() => f(generate(())).generate(()))

    def map[B, I](f: A => I => B): Datagen[B, I] =
      new Datagen.Defer[B, I](input => f(generate(()))(input))

    def map[B](f: A => B): NoInput[B] =
      new NoInput.Defer[B](() => f(generate(())))

  object NoInput:

    final class Defer[A](f: () => A) extends NoInput[A]:
      override def generate(input: Any): A = f()

    final case class Pure[A](a: A) extends NoInput[A]:
      override def generate(input: Any): A = a

  end NoInput

end Datagen
