package gs.datagen.v0

import gs.datagen.v0.generators.Alphabet
import gs.datagen.v0.generators.GenBool
import gs.datagen.v0.generators.GenFiniteDuration
import gs.datagen.v0.generators.GenInteger
import gs.datagen.v0.generators.GenList
import gs.datagen.v0.generators.GenLocalDate
import gs.datagen.v0.generators.GenLong
import gs.datagen.v0.generators.GenMap
import gs.datagen.v0.generators.GenOneOf
import gs.datagen.v0.generators.GenOneOfGen
import gs.datagen.v0.generators.GenOption
import gs.datagen.v0.generators.GenSet
import gs.datagen.v0.generators.GenString
import gs.datagen.v0.generators.GenUUID
import gs.datagen.v0.generators.MinMax
import gs.datagen.v0.generators.Size
import java.time.Clock
import java.time.Instant
import java.time.LocalDate
import java.util.Random
import java.util.UUID
import java.util.concurrent.atomic.AtomicLong
import java.util.concurrent.atomic.AtomicReference
import scala.concurrent.duration.DAYS
import scala.concurrent.duration.FiniteDuration
import scala.concurrent.duration.HOURS
import scala.concurrent.duration.MILLISECONDS
import scala.concurrent.duration.MINUTES
import scala.concurrent.duration.NANOSECONDS
import scala.concurrent.duration.SECONDS
import scala.reflect.ClassTag

/** Type alias for [[Datagen.NoInput]] - a generator which can be executed
  * directly without providing additional input.
  */
type Gen[A] = Datagen.NoInput[A]

/** Provides convenient builders for a number of standard generators. Also
  * controls random data generation and the seeding of that generator.
  */
object Gen:

  private val seed: AtomicLong = new AtomicLong(Instant.now().toEpochMilli())

  private val random: AtomicReference[Random] =
    new AtomicReference(new Random(seed.get()))

  /** Retrieve the current random number generator.
    */
  def rng(): Random = random.get()

  /** Retrieve the current seed being used by generators. This value is useful
    * if generators need to be used in the future with an identical seed. The
    * intended use case is for tests -- the tests run with a seed and fail. By
    * dumping that seed, a re-run can exactly match the original.
    *
    * @return
    *   The currently used seed.
    */
  def getCurrentSeed(): Long = seed.get()

  /** Set the seed. After calling this function, all future calls to generators
    * (including generators which already exist, if they were created via this
    * object or the `generator` package classes) will use this seed.
    *
    * The primary use case of this function is to replay a previous test run
    * using the same seed.
    *
    * Unless a specific seed is needed (such as the replay case), this function
    * should not be used.
    *
    * @param newSeed
    *   The new seed.
    */
  def reseed(newSeed: Long): Unit =
    val _ = seed.set(newSeed)
    val _ = random.set(new Random(newSeed))
    ()

  /** Used as the default clock for dates/times.
    */
  val DefaultClock: Clock = Clock.systemDefaultZone

  /** Generator which produces a single, fixed value.
    *
    * @param value
    *   The fixed value.
    */
  def single[A](value: A): Gen[A] = Datagen.pure(value)

  /** Generator which produces Boolean values.
    *
    * @return
    *   New generator that randomly produces true/false.
    */
  def boolean(): Gen[Boolean] = new GenBool()

  object option:

    /** Generator that will randomly determine whether `Some` or `None` is
      * returned for the given underlying generator.
      *
      * @param generator
      *   The generator that populates the content.
      * @return
      *   The generator for the optional content.
      */
    def apply[A](generator: Gen[A]): Gen[Option[A]] =
      GenOption[A](generator)

    /** Generator that will randomly determine whether `Some` or `None` is
      * returned for the given underlying generator.
      *
      * @param generator
      *   The generator that populates the content.
      * @return
      *   The generator for the optional content.
      */
    def random[A](generator: Gen[A]): Gen[Option[A]] =
      apply(generator)

    /** Generator that will always return `Some`, using some underlying
      * generator.
      *
      * @param generator
      *   The generator that populates the content.
      * @return
      *   The generator for the optional content.
      */
    def alwaysSome[A](generator: Gen[A]): GenOption[A] =
      GenOption.alwaysSome(generator)

    /** Generator that will always return `None`, and therefore will never
      * actually invoke the underlying generator.
      *
      * @param generator
      *   The generator that populates the content.
      * @return
      *   The generator for the optional content.
      */
    def alwaysNone[A](generator: Gen[A]): GenOption[A] =
      GenOption.alwaysNone(generator)

  end option

  /** Generator for a list of some [[Size]] based on a generator for the list
    * elements.
    *
    * @param size
    *   The [[Size]] of the list.
    * @param gen
    *   The generator for the list elements.
    */
  def list[A](
    size: Size,
    gen: Gen[A]
  ): Gen[List[A]] = new GenList[A](size, gen)

  /** Generator for a set of some [[Size]] based on a generator for the set
    * elements.
    *
    * @param size
    *   The goal [[Size]] of the set. If duplicate elements are generated, the
    *   size will be less then specified.
    * @param gen
    *   The generator for the list elements.
    */
  def set[A](
    size: Size,
    gen: Gen[A]
  ): Gen[Set[A]] = new GenSet[A](size, gen)

  /** Generators which pick a single random element from a collection.
    */
  object oneOf:

    /** Generator which randomly selects from a fixed list of choices.
      *
      * This is an alternative to `list` that provides varargs syntax.
      *
      * @param cs
      *   The choices.
      */
    def fixedChoices[A: ClassTag](cs: A*): Gen[A] =
      GenOneOf.list(cs.toList)

    /** Generator which randomly selects from a fixed list of generators on each
      * invocation.
      *
      * @param cs
      *   The choices.
      */
    def generatedChoices[A: ClassTag](cs: Gen[A]*): Gen[A] =
      GenOneOfGen.list(cs.toList)

    /** Generator which randomly selects from a fixed list of choices.
      *
      * @param choices
      *   The choices.
      */
    def fixedList[A: ClassTag](
      choices: List[A]
    ): Gen[A] =
      GenOneOf.list(choices)

    /** Generator which randomly selects from a fixed list of generators on each
      * invocation.
      * @param choices
      *   The choices.
      */
    def generatedList[A: ClassTag](
      choices: List[Gen[A]]
    ): Gen[A] =
      GenOneOfGen.list(choices)

    /** Generator which randomly selects from a fixed set of choices.
      * @param choices
      *   The choices.
      */
    def fixedSet[A: ClassTag](
      choices: Set[A]
    ): Gen[A] =
      GenOneOf.set(choices)

  end oneOf

  /** Generators which build random-sized maps with generated keys and values.
    */
  object map:

    /** Generator for a `Map` based on _separate_ generators for keys and
      * values.
      *
      * @param size
      *   The [[Size]] of the generated map.
      * @param keyGen
      *   The generator for keys.
      * @param valueGen
      *   The generator for values.
      */
    def forKeysAndValues[K, V](
      size: Size,
      keyGen: Gen[K],
      valueGen: Gen[V]
    ): Gen[Map[K, V]] =
      new GenMap.IndependentKeyValue[K, V](
        size = size,
        keyGen = keyGen,
        valueGen = valueGen
      )

    /** Generator for a `Map` based on a _single generator_ that emits tuples of
      * keys to values.
      *
      * @param size
      *   The [[Size]] of the generated map.
      * @param keyValueGen
      *   The generator for key/value pairs.
      */
    def forTuples[K, V](
      size: Size,
      keyValueGen: Gen[(K, V)]
    ): Gen[Map[K, V]] =
      new GenMap.CoupledKeyValue[K, V](
        size = size,
        keyValueGen = keyValueGen
      )

  end map

  /** Generators which produce strings.
    */
  object string:

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted by the given alphabet.
      *
      * @param size
      *   The size constraints for the generated string.
      * @param alphabet
      *   The alphabet from which characters are selected.
      */
    def alphabet(
      size: Size,
      alphabet: Alphabet
    ): Gen[String] =
      new GenString(
        alphabet = alphabet,
        size = size
      )

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted to ASCII lowercase letters.
      *
      * @param size
      *   The size constraints for the generated string.
      */
    def lowercaseAlpha(
      size: Size
    ): Gen[String] =
      new GenString(
        alphabet = Alphabet.ASCII.LowerCaseAlpha,
        size = size
      )

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted to ASCII uppercase letters.
      *
      * @param size
      *   The size constraints for the generated string.
      */
    def uppercaseAlpha(
      size: Size
    ): Gen[String] =
      new GenString(
        alphabet = Alphabet.ASCII.UpperCaseAlpha,
        size = size
      )

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted to ASCII letters.
      *
      * @param size
      *   The size constraints for the generated string.
      */
    def alpha(
      size: Size
    ): Gen[String] =
      new GenString(
        alphabet = Alphabet.ASCII.Alpha,
        size = size
      )

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted to ASCII letters and numbers.
      *
      * @param size
      *   The size constraints for the generated string.
      */
    def alphaNumeric(
      size: Size
    ): Gen[String] =
      new GenString(
        alphabet = Alphabet.ASCII.AlphaNumeric,
        size = size
      )

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted to lowercase ASCII letters and numbers.
      *
      * @param size
      *   The size constraints for the generated string.
      */
    def lowercaseAlphaNumeric(
      size: Size
    ): Gen[String] =
      new GenString(
        alphabet = Alphabet.ASCII.LowerCaseAlphaNumeric,
        size = size
      )

    /** Generator for a string of the specified [[Size]], where the characters
      * are restricted to uppercase ASCII letters and numbers.
      *
      * @param size
      *   The size constraints for the generated string.
      */
    def uppercaseAlphaNumeric(
      size: Size
    ): Gen[String] =
      new GenString(
        alphabet = Alphabet.ASCII.UpperCaseAlphaNumeric,
        size = size
      )

  end string

  /** Generators for integers.
    */
  object integer:

    /** Generator for integers within the specified (inclusive) range.
      *
      * @param lowerBound
      *   The lowest possible value that will be generated.
      * @param upperBound
      *   the highest possible value that will be generated.
      */
    def inRange(
      lowerBound: Int,
      upperBound: Int
    ): Gen[Int] =
      GenInteger.inRange(lowerBound, upperBound)

    /** Generator for positive (greater than 0) integers.
      */
    def positive(): Gen[Int] =
      GenInteger.positive()

    /** Generator for negative (less than 0) integers.
      */
    def negative(): Gen[Int] =
      GenInteger.negative()

    /** Generator for non-negative (0 or greater) integers.
      */
    def nonNegative(): Gen[Int] =
      GenInteger.nonNegative()

  end integer

  /** Generators for longs.
    */
  object long:

    /** Generator for longs within the specified (inclusive) range.
      *
      * @param lowerBound
      *   The lowest possible value that will be generated.
      * @param upperBound
      *   the highest possible value that will be generated.
      */
    def inRange(
      lowerBound: Long,
      upperBound: Long
    ): Gen[Long] =
      GenLong.inRange(lowerBound, upperBound)

    /** Generator for positive (greater than 0) longs.
      */
    def positive(): Gen[Long] =
      GenLong.positive()

    /** Generator for negative (less than 0) longs.
      */
    def negative(): Gen[Long] =
      GenLong.negative()

    /** Generator for non-negative (0 or greater) longs.
      */
    def nonNegative(): Gen[Long] =
      GenLong.nonNegative()

  end long

  /** Generators for dates (the `java.time.LocalDate` type).
    */
  object date:

    /** Generate a date before the given date.
      *
      * ## Example
      *
      * The following generator produces a date before today where:
      *
      *   - The day component is subtracted by 0 to 10.
      *   - The month component is subtracted by 1 to 11.
      *   - The year component is subtracted by 0 to 30.
      *
      * The produced date will be, at most, 30 years, 11 months and 10 days
      * prior to the selected pivot. It will be at least 1 month prior to the
      * selected pivot.
      *
      * {{{
      * Gen.date.before(
      *   date = LocalDate.now(),
      *   days = MinMax.nonNegative(0, 10),
      *   months = MinMax.nonNegative(1, 11),
      *   years = MinMax.nonNegative(0, 30)
      * )
      * }}}
      *
      * @param date
      *   The maximum date.
      * @param days
      *   The range of days prior to the given date.
      * @param months
      *   The range of months prior to the given date.
      * @param years
      *   The range of years prior to the given date.
      */
    def before(
      date: LocalDate,
      days: MinMax.NonNegative,
      months: MinMax.NonNegative = MinMax.Zero,
      years: MinMax.NonNegative = MinMax.Zero
    ): Gen[LocalDate] =
      new GenLocalDate.Before(
        pivot = date,
        days = days,
        months = months,
        years = years
      )

    def beforeToday(
      days: MinMax.NonNegative,
      months: MinMax.NonNegative = MinMax.Zero,
      years: MinMax.NonNegative = MinMax.Zero,
      clock: Clock = DefaultClock
    ): Gen[LocalDate] =
      new GenLocalDate.Before(
        pivot = LocalDate.now(clock),
        days = days,
        months = months,
        years = years
      )

    def after(
      date: LocalDate,
      days: MinMax.NonNegative,
      months: MinMax.NonNegative = MinMax.Zero,
      years: MinMax.NonNegative = MinMax.Zero
    ): Gen[LocalDate] =
      new GenLocalDate.After(
        pivot = date,
        days = days,
        months = months,
        years = years
      )

    def afterToday(
      days: MinMax.NonNegative,
      months: MinMax.NonNegative = MinMax.Zero,
      years: MinMax.NonNegative = MinMax.Zero,
      clock: Clock = DefaultClock
    ): Gen[LocalDate] =
      new GenLocalDate.After(
        pivot = LocalDate.now(clock),
        days = days,
        months = months,
        years = years
      )

    def around(
      date: LocalDate,
      days: MinMax,
      months: MinMax = MinMax.Zero,
      years: MinMax = MinMax.Zero
    ): Gen[LocalDate] =
      new GenLocalDate.After(
        pivot = date,
        days = days,
        months = months,
        years = years
      )

    def aroundToday(
      days: MinMax,
      months: MinMax = MinMax.Zero,
      years: MinMax = MinMax.Zero,
      clock: Clock = DefaultClock
    ): Gen[LocalDate] =
      new GenLocalDate.After(
        pivot = LocalDate.now(clock),
        days = days,
        months = months,
        years = years
      )

  end date

  /** Geneators for `java.util.UUID` values.
    */
  object uuid:

    /** Create a new UUID generator that emits type-4 (random) UUIDs.
      */
    def random(): Gen[UUID] = new GenUUID

    /** Generator for UUIDs represented as strings.
      */
    def string(): Gen[String] = uuid.random().map(id => id.toString)

    /** Generator for UUIDs represented as strings, without dashes.
      */
    def noDashes(): Gen[String] =
      uuid.random().map(id => id.toString.replace("-", ""))

  end uuid

  /** Generators for `scala.concurrent.duration._` types such as
    * `FiniteDuration`.
    */
  object duration:

    /** Generator for a duration expressed in nanoseconds for some finite range.
      *
      * @param lowerBound
      *   The lower bound on the generated number of nanoseconds.
      * @param upperBound
      *   The upper bound on the generated number of nanoseconds.
      */
    def finiteNanoseconds(
      lowerBound: Long,
      upperBound: Long
    ): Gen[FiniteDuration] =
      GenFiniteDuration(lowerBound, upperBound, NANOSECONDS)

    /** Generator for a duration expressed in milliseconds for some finite
      * range.
      *
      * @param lowerBound
      *   The lower bound on the generated number of milliseconds.
      * @param upperBound
      *   The upper bound on the generated number of milliseconds.
      */
    def finiteMilliseconds(
      lowerBound: Long,
      upperBound: Long
    ): Gen[FiniteDuration] =
      GenFiniteDuration(lowerBound, upperBound, MILLISECONDS)

    /** Generator for a duration expressed in seconds for some finite range.
      *
      * @param lowerBound
      *   The lower bound on the generated number of seconds.
      * @param upperBound
      *   The upper bound on the generated number of seconds.
      */
    def finiteSeconds(
      lowerBound: Long,
      upperBound: Long
    ): Gen[FiniteDuration] =
      GenFiniteDuration(lowerBound, upperBound, SECONDS)

    /** Generator for a duration expressed in minutes for some finite range.
      *
      * @param lowerBound
      *   The lower bound on the generated number of minutes.
      * @param upperBound
      *   The upper bound on the generated number of minutes.
      */
    def finiteMinutes(
      lowerBound: Long,
      upperBound: Long
    ): Gen[FiniteDuration] =
      GenFiniteDuration(lowerBound, upperBound, MINUTES)

    /** Generator for a duration expressed in hours for some finite range.
      *
      * @param lowerBound
      *   The lower bound on the generated number of hours.
      * @param upperBound
      *   The upper bound on the generated number of hours.
      */
    def finiteHours(
      lowerBound: Long,
      upperBound: Long
    ): Gen[FiniteDuration] =
      GenFiniteDuration(lowerBound, upperBound, HOURS)

    /** Generator for a duration expressed in days for some finite range.
      *
      * @param lowerBound
      *   The lower bound on the generated number of days.
      * @param upperBound
      *   The upper bound on the generated number of days.
      */
    def finiteDays(
      lowerBound: Long,
      upperBound: Long
    ): Gen[FiniteDuration] =
      GenFiniteDuration(lowerBound, upperBound, DAYS)

  end duration

end Gen
