package gs.datagen.v0.generators

import java.lang.StringBuilder
import java.util.Random

/** Represents some set of characters which can be generated.
  */
abstract class Alphabet:
  def next(random: Random): String
  def nextChar(random: Random): Char

  def fill(
    random: Random,
    count: Int
  ): String

object Alphabet:

  /** Represents a character range based on a lower and upper integer bound
    * (inclusive).
    */
  final class CharacterRange private (
    val lowerBound: Int,
    val upperBound: Int
  ):

    override def equals(obj: Any): Boolean =
      obj match
        case other: CharacterRange =>
          lowerBound == other.lowerBound && upperBound == other.upperBound
        case _ => false

    override def hashCode(): Int = lowerBound * 31 + upperBound

    override def toString(): String = s"[$lowerBound, $upperBound]"

  object CharacterRange:

    /** Represents lower case ASCII alphabetical characters `[a-z]`.
      */
    val LowerCaseAlpha: CharacterRange = apply(97, 122)

    /** Represents upper case ASCII alphabetical characters `[A-Z]`.
      */
    val UpperCaseAlpha: CharacterRange = apply(65, 90)

    /** Represents numeic ASCII characters `[0-9]`.
      */
    val Numeric: CharacterRange = apply(48, 57)

    def apply(
      lower: Int,
      upper: Int
    ): CharacterRange =
      if lower < 0 then
        throw new IllegalArgumentException(
          s"The lower bound of a character range must be greater than or equal to 0. Received: '$lower'"
        )
      else if lower > upper then
        throw new IllegalArgumentException(
          s"The lower bound of a character range must be less than or equal to the upper bound. Received lower bound '$lower' and upper bound '$upper'"
        )
      else new CharacterRange(lower, upper)

  end CharacterRange

  /** Alphabet that consists of a single character.
    *
    * @param c
    *   The character.
    */
  final class SingleCharacter(c: Char) extends Alphabet:
    override def next(random: Random): String   = c.toString
    override def nextChar(random: Random): Char = c

    override def fill(
      random: Random,
      count: Int
    ): String =
      if count > 0 then c.toString * count
      else ""

  /** Alphabet that consists of a single [[CharacterRange]]. Generated
    * characters are randomly selected from this range.
    */
  class SingleRange(
    range: CharacterRange
  ) extends Alphabet:

    override def next(random: Random): String =
      nextChar(random).toString

    override def nextChar(random: Random): Char =
      random.nextInt(range.lowerBound, range.upperBound + 1).toChar

    override def fill(
      random: Random,
      count: Int
    ): String =
      if count > 0 then
        val sb = new StringBuilder
        (1 to count).foreach(_ => sb.append(nextChar(random)))
        sb.toString
      else ""

  /** Alphabet that consists of multiple unique [[CharacterRange]]. Generated
    * characters are randomly distributed across all ranges. At least one range
    * _must_ be specified.
    */
  class MultiRange protected (
    ranges: Array[CharacterRange]
  ) extends Alphabet:

    override def next(random: Random): String =
      nextChar(random).toString

    override def nextChar(random: Random): Char =
      charInRange(random, ranges.apply(random.nextInt(ranges.size)))

    private def charInRange(
      random: Random,
      range: CharacterRange
    ): Char =
      random.nextInt(range.lowerBound, range.upperBound + 1).toChar

    override def fill(
      random: Random,
      count: Int
    ): String =
      if count > 0 then
        val sb = new StringBuilder
        (1 to count).foreach(_ => sb.append(nextChar(random)))
        sb.toString
      else ""

  object MultiRange:

    def apply(ranges: Set[CharacterRange]): MultiRange =
      if ranges.isEmpty then
        throw new IllegalArgumentException(
          "At least one range must be specified."
        )
      else new MultiRange(ranges.toArray)

  end MultiRange

  /** Predefined ranges for the ASCII character set.
    */
  object ASCII:

    /** Alphabet for `[a-z]`
      */
    val LowerCaseAlpha: Alphabet = SingleRange(CharacterRange.LowerCaseAlpha)

    /** Alphabet for `[A-Z]`
      */
    val UpperCaseAlpha: Alphabet = SingleRange(CharacterRange.UpperCaseAlpha)

    /** Alphabet for `[a-zA-Z]`
      */
    val Alpha: Alphabet = MultiRange(
      Set(CharacterRange.LowerCaseAlpha, CharacterRange.UpperCaseAlpha)
    )

    /** Alphabet for `[0-9]`
      */
    val Numeric: Alphabet = SingleRange(CharacterRange.Numeric)

    /** Alphabet for `[a-zA-Z0-9]`
      */
    val AlphaNumeric: Alphabet = MultiRange(
      Set(
        CharacterRange.LowerCaseAlpha,
        CharacterRange.UpperCaseAlpha,
        CharacterRange.Numeric
      )
    )

    /** Alphabet for `[a-z0-9]`
      */
    val LowerCaseAlphaNumeric: Alphabet = MultiRange(
      Set(
        CharacterRange.LowerCaseAlpha,
        CharacterRange.Numeric
      )
    )

    /** Alphabet for `[A-Z0-9]`
      */
    val UpperCaseAlphaNumeric: Alphabet = MultiRange(
      Set(
        CharacterRange.UpperCaseAlpha,
        CharacterRange.Numeric
      )
    )

  end ASCII

end Alphabet
