package gs.datagen.v0.generators

import java.util.Random

/** Represents a variable size which must (usually) be greater than or equal to
  * 0.
  */
abstract class Size:
  def next(random: Random): Int

object Size:

  def fixed(size: Int): Size = Fixed(size)

  def between(
    lower: Int,
    upper: Int
  ): Size = Between(lower, upper)

  final class Fixed private (
    size: Int
  ) extends Size:
    override def next(random: Random): Int = size

  object Fixed:

    def apply(
      size: Int
    ): Size =
      if size < 0 then
        throw new IllegalArgumentException(
          s"The size must be greater than or equal to 0. Received: '$size'"
        )
      else new Fixed(size)

  end Fixed

  final class Between private (
    lowerBound: Int,
    upperBound: Int
  ) extends Size:

    override def next(random: Random): Int =
      random.nextInt(lowerBound, upperBound + 1)

  object Between:

    def apply(
      lowerBound: Int,
      upperBound: Int
    ): Size =
      if lowerBound < 0 then
        throw new IllegalArgumentException(
          s"The lower bound of a size must be greater than or equal to 0. Received: '$lowerBound'"
        )
      else if lowerBound > upperBound then
        throw new IllegalArgumentException(
          s"The lower bound of a size must less than the upper bound. Received lowerBound='$lowerBound' and upperBound='$upperBound'"
        )
      else new Between(lowerBound, upperBound)

  end Between

end Size
